/* eslint-disable no-unused-vars */
const REL_PATH_DATA = "";
const URL_DOMAIN = "http://127.0.0.1";
const URL_DEFAULT_PORT = "8888";
const URL_HOST = URL_DOMAIN + ":" + URL_DEFAULT_PORT;
const URL_PATH_UPDATE_XML = "/" + REL_PATH_DATA + "app_update.sjs";
const URL_HTTP_UPDATE_SJS = URL_HOST + URL_PATH_UPDATE_XML;
const CONTINUE_CHECK = "continueCheck";
const CONTINUE_DOWNLOAD = "continueDownload";
const CONTINUE_STAGING = "continueStaging";
