/* -*- indent-tabs-mode: nil; js-indent-level: 2 -*- */
/* vim:set ts=2 sw=2 sts=2 et: */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

// Regression test for bug 333423 - crash enumerating entries of a
// closed nsIZipReader
function run_test() {
  // the build script have created the zip we can test on in the current dir.
  var file = do_get_file("data/test_bug333423.zip");

  var zipreader = Cc["@mozilla.org/libjar/zip-reader;1"].createInstance(
    Ci.nsIZipReader
  );
  zipreader.open(file);
  zipreader.close();
  // this should error out and not crash
  Assert.throws(
    () => zipreader.findEntries("*.*"),
    /NS_ERROR_FAILURE/,
    "Should error out on a closed zipreader"
  );
}
